import pytest
import numpy as np
from pydif import pydif 
from pydif.pydif import diff, diffdiff
from pydif.elementary import elementary as el


def test_init():
    def f1(x):
        return x
    def f2(x, y):
        return x+y

    ad1 = pydif.autodiff(f1)
    ad2 = pydif.autodiff(f2)

    assert(ad1.func == f1)
    assert(ad2.func == f2)
    assert(ad1.num_params == 1)
    assert(ad2.num_params == 2)

def test_check_dim():
    def f1(x):
        return x
    def f2(x, y):
        return x+y

    ad1 = pydif.autodiff(f1)
    ad2 = pydif.autodiff(f2)

    with pytest.raises(ValueError):
        ad1._check_dim([0,0])
    ad1._check_dim([0])
    ad1._check_dim(0)

    with pytest.raises(ValueError):
        ad2._check_dim([0])
    with pytest.raises(ValueError):
        ad2._check_dim(0)
    ad2._check_dim([0,0])

def test_enforce_unitvector():
    def f1(x):
        return x

    dir1 = np.array([1])
    dir2 = np.array([0.5])
    dir3 = np.array([0,1])
    dir4 = np.array([1,-1])
    dir5 = np.array([0,0])

    ad1 = pydif.autodiff(f1)

    assert(np.isclose(ad1._enforce_unitvector(dir1),np.array([1])))
    assert(np.isclose(ad1._enforce_unitvector(dir2),np.array([1])))
    assert(all(np.isclose(ad1._enforce_unitvector(dir3),np.array([0,1]))))
    assert(all(np.isclose(ad1._enforce_unitvector(dir4),np.array([np.sqrt(1/2),-np.sqrt(1/2)]))))

    with pytest.raises(ValueError):
        ad1._enforce_unitvector(dir5)


def test_eval():
    def f1(x):
        return x
    def f2(x, y):
        return x+y

    ad1 = pydif.autodiff(f1)
    ad2 = pydif.autodiff(f2)

    pos1 = 3
    pos2 = [3,4]

    assert(ad1._eval(pos1, wrt_variables=False).val == pos1)
    assert(ad1._eval(pos1, wrt_variables=False).der == 1)
    assert(ad1._eval(pos1, wrt_variables=False).der2 == 0)
    assert(ad1._eval(pos1, wrt_variables=True).val == pos1)
    assert(ad1._eval(pos1, wrt_variables=True).der == 1)
    assert(ad1._eval(pos1, wrt_variables=True).der2 == 0)

    assert(ad2._eval(pos2, wrt_variables=False).val == pos2[0] + pos2[1])
    assert(ad2._eval(pos2, wrt_variables=False).der == 2)
    assert(ad2._eval(pos2, wrt_variables=False).der2 == 0)
    assert(ad2._eval(pos2, wrt_variables=True).val == pos2[0] + pos2[1])
    assert(all(ad2._eval(pos2, wrt_variables=True).der == [1,1]))
    assert(all(ad2._eval(pos2, wrt_variables=True).der2 == [0,0]))

def test_get_val():
    def f1(x):
        return x
    def f2(x, y):
        return x, y

    ad1 = pydif.autodiff(f1)
    ad2 = pydif.autodiff(f2)

    pos1 = 3
    pos2 = [3,4]
    dir1 = [1]
    dir2 = [0,1]

    assert(ad1.get_val(pos1, direction=None) == pos1)
    assert(ad1.get_val(pos1, direction=dir1) == pos1)

    assert(ad2.get_val(pos2, direction=None) == pos2)
    assert(ad2.get_val(pos2, direction=dir2) == pos2[1])

def test_get_val_vector():
    def f1(x):
        return x
    def f2(x, y):
        return x, y

    ad1 = pydif.autodiff_vector([[f1, f2], [f2, f1]])
    ad2 = pydif.autodiff_vector([f2, f2, f2, f2])

    pos1 = 3
    pos2 = [3,4]
    dir1 = [1]
    dir2 = [0,1]
    dir3 = [1,0]

    res = ad1.get_val(pos2, direction=None)
    assert(res[0][0] == res[1][1] == pos1)
    assert(res[0][1] == res[1][0] == pos2)

    res = ad1.get_val(pos2, direction=dir3)
    assert(res[0][0] == res[1][1] == pos1)
    assert(res[0][1] == res[1][0] == pos2[0])

    res = ad2.get_val(pos2, direction=None)
    assert(res[0] == res[1] == res[2] == res[3] == pos2)

    res = ad2.get_val(pos2, direction=dir2)
    assert(res[0] == res[1] == res[2] == res[3] == pos2[1])

def test_get_der():
    def f1(x):
        return x
    def f2(x, y):
        return x, y

    ad1 = pydif.autodiff(f1)
    ad2 = pydif.autodiff(f2)

    pos1 = 3
    pos2 = [3,4]
    dir1 = [1]
    dir2 = [0,1]

    assert(ad1.get_der(pos1, direction=None) == 1)
    assert(ad1.get_der(pos1, direction=dir1) == 1)

    assert(ad2.get_der(pos2, direction=None) == [1,1])
    assert(ad2.get_der(pos2, direction=dir2) == 1)

def test_get_der_vector():
    def f1(x):
        return x
    def f2(x, y):
        return x, y

    ad1 = pydif.autodiff_vector([[f1, f2], [f2, f1]])
    ad2 = pydif.autodiff_vector([f2, f2, f2, f2])

    pos1 = 3
    pos2 = [3,4]
    dir1 = [1]
    dir2 = [0,1]
    dir3 = [1,0]

    res = ad1.get_der(pos2, direction=None)
    assert(res[0][0] == res[1][1] == 1)
    assert(res[0][1] == res[1][0] == [1,1])

    res = ad1.get_der(pos2, direction=dir3)
    assert(res[0][0] == res[1][1] == 1)
    assert(res[0][1] == res[1][0] == 1)

    res = ad2.get_der(pos2, direction=None)
    assert(res[0] == res[1] == res[2] == res[3] == [1,1])

    res = ad2.get_der(pos2, direction=dir2)
    assert(res[0] == res[1] == res[2] == res[3] == 1)

def test_get_der2():
    def f1(x):
        return x
    def f2(x, y):
        return x, y

    ad1 = pydif.autodiff(f1)
    ad2 = pydif.autodiff(f2)

    pos1 = 3
    pos2 = [3,4]
    dir1 = [1]
    dir2 = [0,1]

    assert(ad1.get_der(pos1, direction=None, order=2) == 0)
    assert(ad1.get_der(pos1, direction=dir1, order=2) == 0)

    assert(ad2.get_der(pos2, direction=None, order=2) == [0,0])
    assert(ad2.get_der(pos2, direction=dir2, order=2) == 0)

def test_multiply_add_simple():
    alpha = 2
    pos = 2

    def f1(x):
        return alpha*x+3
    def f2(x):
        return x*alpha+3
    def f3(x):
        return 3+x*alpha
    def f4(x):
        return 3+alpha*x

    ad1 = pydif.autodiff(f1)
    ad2 = pydif.autodiff(f2)
    ad3 = pydif.autodiff(f3)
    ad4 = pydif.autodiff(f4)

    assert(ad1.get_val(pos) == 7)
    assert(ad2.get_val(pos) == 7)
    assert(ad3.get_val(pos) == 7)
    assert(ad4.get_val(pos) == 7)

    assert(ad1.get_der(pos) == 2)
    assert(ad2.get_der(pos) == 2)
    assert(ad3.get_der(pos) == 2)
    assert(ad4.get_der(pos) == 2)

    assert(ad1.get_der(pos, order=2) == 0)
    assert(ad1.get_der(pos, order=2) == 0)
    assert(ad1.get_der(pos, order=2) == 0)
    assert(ad1.get_der(pos, order=2) == 0)

def test_divide_add_simple():
    alpha = 2
    pos = 2

    def f1(x):
        return alpha/x+3
    def f2(x):
        return 3+alpha/x
    def f3(x):
        return x/alpha+3
    def f4(x):
        return 3+x/alpha

    ad1 = pydif.autodiff(f1)
    ad2 = pydif.autodiff(f2)
    ad3 = pydif.autodiff(f3)
    ad4 = pydif.autodiff(f4)

    assert(ad1.get_val(pos) == 4)
    assert(ad2.get_val(pos) == 4)
    assert(ad3.get_val(pos) == 4)
    assert(ad4.get_val(pos) == 4)

    assert(ad1.get_der(pos) == -0.5)
    assert(ad2.get_der(pos) == -0.5)
    assert(ad3.get_der(pos) == 0.5)
    assert(ad4.get_der(pos) == 0.5)

    assert(ad1.get_der(pos, order=2) == 0.5)
    assert(ad2.get_der(pos, order=2) == 0.5)
    assert(ad3.get_der(pos, order=2) == 0)
    assert(ad4.get_der(pos, order=2) == 0)

def test_composite():
    pos = (1,2,3)

    def f1(x,y,z):
        return (1/(x*y*z)) + el.sin((1/x) + (1/y) + (1/z))
    def f2(x,y,z):
        return (1/(x*y*z))

    ad1 = pydif.autodiff(f1)
    ad2 = pydif.autodiff(f2)

    expected_val = (1/6) + np.sin(11/6)

    assert(ad1.get_val(pos) == expected_val)

    der1 = -(1/6) - np.cos(11/6) #d/dx
    der2 = -(1/12) - (np.cos(11/6)/4) #d/dy
    der3 = -(1/18) - (np.cos(11/6)/9) #d/dz

    assert(np.allclose(ad1.get_der(pos,wrt_variables=True),[der1,der2,der3]))

    assert(ad2.get_val(pos) == 1/(1*2*3))
    #http://www.wolframalpha.com/input/?i=d%2Fda+(1%2F(x(a)*y(a)*z(a))
    expected_der = -((2*3 + 1*3 + 1*2)/((1**2) * (2**2) * (3**2)))
    assert(np.isclose(float(ad2.get_der(pos)), expected_der))

    der2_1 = 2/(1*2*3) #d/dx^2
    der2_2 = 2/(1*((2)**3)*3) #d/dy^2
    der2_3 = 2/(1*2*((3)**3)) #d/dz^2
    assert(np.allclose(ad2.get_der(pos,wrt_variables=True,order=2),[der2_1, der2_2, der2_3]))

def test_derorder():
    def f1(x):
        return x

    ad1 = pydif.autodiff(f1)

    assert(ad1.get_der(2, order=1) == 1)
    assert(ad1.get_der(2, order=2) == 0)
    with pytest.raises(NotImplementedError):
        ad1.get_der(2, order=3)

def test_decorator():
    @diff
    def f1d(x):
        return x
    @diffdiff
    def f1dd(x):
        return x

    alpha=2

    @diff
    def f2d(x):
        return alpha*x+3
    @diffdiff
    def f2dd(x):
        return alpha*x+3

    pos = (1,2,3)

    @diff
    def f3d(x,y,z):
        return (1/(x*y*z)) + el.sin((1/x) + (1/y) + (1/z))

    assert(f1d(3) == 1)
    assert(f1dd(3) == 0)

    assert(f2d(2) == 2)
    assert(f2dd(2) == 0)

    der1 = -(1/6) - np.cos(11/6) #d/dx
    der2 = -(1/12) - (np.cos(11/6)/4) #d/dy
    der3 = -(1/18) - (np.cos(11/6)/9) #d/dz

    assert(np.allclose(f3d(pos),[der1,der2,der3]))



import sys
import os
import numpy as np
from inspect import signature
from collections import Iterable
sys.path.append(os.path.join(os.getcwd(),'pydif'))
from pydif.pydif import autodiff

class Optimize():
    def __init__(self, func):
        self.func = func

    def gradient_descent(self, init_pos, step_size=0.1, max_iters=100, precision=0.001, return_hist=False):

        num_params = len(signature(self.func).parameters)
        badDimentionsMsg = 'poorly formatted initial position. should be of length {}.'.format(num_params)
        if num_params != len(init_pos):
            raise ValueError(badDimentionsMsg)

        cur_pos = init_pos
        iters = 0
        dfdx = autodiff(self.func)
        val = dfdx.get_val(init_pos)

        if isinstance(val, Iterable):
            raise ValueError("The optimize class only optimizes scalar valued functions")

        # pre allocate history array
        hist = [cur_pos]
        prev_step_size = 100 + precision
        while (prev_step_size > precision and iters < max_iters):
            jac = dfdx.get_der(cur_pos, wrt_variables=True)
            prev_pos = cur_pos
            cur_pos = cur_pos - step_size * jac
            prev_step_size = np.linalg.norm(abs(cur_pos - prev_pos))
            iters += 1
            hist.append(cur_pos) #store history
        np.array(hist)

        if return_hist:
            return cur_pos, hist
        else:
            return cur_pos

    #define steepest descent
    def steepest(self, init_pos, step_size=0.1, max_iters=100, precision=0.001, return_hist=False):

        num_params = len(signature(self.func).parameters)
        badDimentionsMsg = 'poorly formatted initial position. should be of length {}.'.format(num_params)

        if num_params != len(init_pos):
            raise ValueError(badDimentionsMsg)

        cur_pos = init_pos
        iters = 0
        dfdx = autodiff(self.func)
        val = dfdx.get_val(init_pos)

        if isinstance(val, Iterable):
            raise ValueError("The optimize class only optimizes scalar valued functions")

        #preallocate arrays
        hist = [cur_pos]

        #set inital conditions
        iters = 0
        s = 0
        step = 100000

        #define conditions to break loop
        while ((iters <= max_iters) and (step >= precision)):
            s = -dfdx.get_val(cur_pos) #step
            n = line_search(f, grad_f, cur_pos, s)[0] #TODO replace line_search
            step = np.linalg.norm(n*s) #set step size
            cur_pos = cur_pos + n * s #append value after step
            iters += 1 #count step
            hist.append(cur_pos) #store history
        np.array(hist)

        if return_hist:
            return cur_pos, hist
        else:
            return cur_pos

    def newton(self, init_pos, step_size=0.1, max_iters=100, precision=0.001, return_hist=False):

        num_params = len(signature(self.func).parameters)
        badDimentionsMsg = 'poorly formatted initial position. should be of length {}.'.format(num_params)

        if num_params != len(init_pos):
            raise ValueError(badDimentionsMsg)

        cur_pos = init_pos
        iters = 0
        dfdx = autodiff(self.func)
        val = dfdx.get_val(init_pos)

        if isinstance(val, Iterable):
            raise ValueError("The optimize class only optimizes scalar valued functions")

        #preallocate arrays
        hist = [cur_pos]

        #set inital conditions
        iters = 0
        s_k = 0
        step = 10000

        #TODO make hessian a function or change the call in the line below
        hessian = dfdx.get_der(cur_pos, wrt_variables=True, order = 2)

        #define conditions to break loop
        while ((iters <= max_iters) and (step >= precision)):
            s = np.linalg.solve(hessian(cur_pos), -dfdx.get_val(cur_pos)) #TODO make hessian work
            step = np.linalg.norm(s) # get step size
            cur_pos = cur_pos + s #step
            iters += 1 #increment counter
            hist.append(cur_pos) #store history
        hist = np.array(hist)

        if return_hist:
            return cur_pos, hist
        else:
            return cur_pos

    #function that allows for numerous initial conditions to be specified and plotted
    def plot_optimization(optimizer, initial_cond):

        #define initial conditions and plot results
        for init in initial_cond:
            min_val, total_it, hist = self.optimizer(init[0], init[1]) #call optimization function
            hist = np.array(hist)

            #format plot
            xs = np.linspace(-5,5,1000)
            ys = np.linspace(-5,5,1000)
            X, Y = np.meshgrid(xs, ys)
            Z = f(np.array([X,Y])) # TODO f is the function we  are optimizing
            plt.contour(X, Y, Z, 100)
            plt.plot(hist[:,0], hist[:,1])
            plt.ylim((ys[0],ys[-1]))
            plt.xlim((xs[0],xs[-1]))
            plt.xlabel('x')
            plt.ylabel('y')
            plt.show()



"""dual.py
This file overloads operators dunder methods to allow for dual numbers to be used. The functions first try to work
with x as a dual number and falls back to treating x as a normal numerical type
Args:
    param1 (float): The value of a function at a point.
    param2 (array or array-like): An array of partial first derivatives at a point
    param2 (array or array-like): An array of partial second derivatives at a point

Returns:
    Dual Number Object: Returns a new Dual number object with the above parameters.

"""

import numpy as np
class Dual():

    #initialize derivative as a numpy array to handle partial derivatives
    def __init__(self, val , der, der2):
        self.val = val
        self.der = np.array(der)
        self.der2 = np.array(der2)

    #overload the addition method
    def __add__(self, x):
        try:
            return Dual(self.val +  x.val, self.der +  x.der, self.der2 +  x.der2)
        except AttributeError:
            return Dual(self.val +  x, self.der, self.der2)

    #overload radd by calling addition on the Dual number
    def __radd__(self, x):
        return self.__add__(x)

    #overload the subtraction method
    def __sub__(self, x):
        try:
            return Dual(self.val -  x.val, self.der -  x.der, self.der2 -  x.der2)
        except AttributeError:
            return Dual(self.val -  x, self.der, self.der2)

    #overload rsub
    def __rsub__(self, x):
        try:
            return Dual(self.val -  x.val, self.der -  x.der, self.der2 -  x.der2)
        except AttributeError:
            return Dual(x - self.val, self.der, self.der2)

    #overload multiplication
    def __mul__(self, x):
        try:
            return Dual(self.val * x.val, self.der * x.val + self.val * x.der, x.val * self.der2 + 2 * self.der * x.der + self.der * x.der2)
        except AttributeError:
            return Dual(self.val *  x, self.der *  x, self.der2 *  x)

    #overload rmul by calling multiplication on the Dual number
    def __rmul__(self, x):
        return self.__mul__(x)

    #overload division
    def __truediv__(self, x):
        try:
            return Dual(self.val/ x.val, (self.der * x.val - self.val * x.der)/(x.val)**2, (x.val ** 2 * self.der2 - x.val * (2 * self.der * x.der + self.val * x.der2) + 2 * self.val * x.der **2)/x.val**3)
        except AttributeError:
            return Dual(self.val /x, self.der / x, self.der2 / x)

    #overload rdivision by multiplying by the value to the negative first power
    def __rtruediv__(self, x):
        return Dual(x/self.val, -(x * self.der)/self.val ** 2, (2 * x * self.der ** 2 - x * self.val * self.der2 )/self.val**3)

    #overload power operator using formula for derivative of a function raised to a function if both are dual numbers
    def __pow__(self, x):
        try:
            return Dual(self.val**x.val, self.val**x.val*(self.der*(x.val/self.val)+x.der * np.log(self.val)), self.val ** x.val * ((x.val * self.der)/self.val + np.log(self.val) * x.der) ** 2 + self.val ** x.val * ((x.val * self.der2)/self.val + (2 * self.der * x.der)/self.val - (x.val * self.der**2)/self.val**2 + np.log(self.val) * x.der2))
        except AttributeError:
            return Dual(x**self.val, np.log(x) * x ** self.val  * self.der, np.log(x) * x  ** self.val * (self.der2 + np.log(x) * self.der  ** 2))

    #overload rpow similarly to above
    def __rpow__(self, x):
        return Dual(x**self.val, np.log(x) * x ** self.val  * self.der, np.log(x) * x  ** self.val * (self.der2 + np.log(x) * self.der  ** 2))

    #overload negation
    def __neg__(self):
        return Dual(-self.val, -self.der, -self.der2)

    #overload equality
    def __eq__(self, x):
        try:
            return (self.val == x.val and np.array_equal(self.der, x.der) and np.array_equal(self.der2, x.der2))
        except: #if not Dual, compare to value
            return self.val == x

    #overload inequality
    def __neq__(self, x):
        try:
            return (self.val != x.val or (np.array_equal(self.der, x.der) == False) or (np.array_equal(self.der2, x.der2) == False))
        except: #if not Dual, compare to value
            return (self.val != x)

    #overload repr by displaying as a list where the first value is the value and the second is a derivative
    def __repr__(self):
        return '[{0}, {1}, {2}]'.format(self.val, self.der, self.der2)



from pydif.elementary import elementary as el
import pytest
# from ..dual.dual import Dual
from pydif.dual.dual import Dual

def test_cos():
    assert(el.cos(5) == pytest.approx(0.2836, 0.001))
    x = Dual(5, 8, 1)
    cosx = el.cos(x)
    assert(cosx.val == pytest.approx(0.2836, 0.001))
    assert(cosx.der == pytest.approx(7.671, 0.001))
    assert(cosx.der2 == pytest.approx(-17.1954, 0.001))

def test_sin():
    assert(el.sin(3) == pytest.approx(0.1411, 0.001))
    x = Dual(3, 3, 1)
    sinx = el.sin(x)
    assert(sinx.val == pytest.approx(0.1411, 0.001))
    assert(sinx.der == pytest.approx(-2.9699, 0.001))
    assert(sinx.der2 == pytest.approx(-2.260, 0.001))

def test_tan():
    assert(el.tan(2) == pytest.approx(-2.185, 0.001))
    x = Dual(2, 3, 4)
    tanx = el.tan(x)
    assert(tanx.val == pytest.approx(-2.185, 0.001))
    assert(tanx.der == pytest.approx(17.32319, 0.001))
    assert(tanx.der2 == pytest.approx(-204.0136, 0.001))

def test_exp():
    assert(el.exp(0) == pytest.approx(1))
    assert(el.exp(5)) == pytest.approx(148.413159, 0.0001)
    x = Dual(3, 4, 5)
    ex = el.exp(x)
    assert(ex.val == pytest.approx(20.0855369, 0.0001))
    assert(ex.der == pytest.approx(80.3421476, 0.0001))
    assert(ex.der2 == pytest.approx(421.7962, 0.001))

def test_arcsin():
    assert(el.arcsin(1) == pytest.approx(1.570, 0.001))
    x = Dual(0.5, 2, 2)
    arcsinx = el.arcsin(x)
    assert(arcsinx.val == pytest.approx(0.5235, 0.01))
    assert(arcsinx.der == pytest.approx(1.154*2, 0.01))
    assert(arcsinx.der2 == pytest.approx(5.3886, 0.01))

def test_arccos():
    assert(el.arccos(0.9) == pytest.approx(0.451, 0.001))
    x = Dual(0.5, 2, 4)
    arccosx = el.arccos(x)
    assert(arccosx.val == pytest.approx(1.047, 0.01))
    assert(arccosx.der == pytest.approx( -2 *1.1547, 0.01))
    assert(arccosx.der2 == pytest.approx(-7.698, 0.01))

def test_arctan():
    assert(el.arctan(0.75) == pytest.approx(0.6435, 0.001))
    x = Dual(0.5, 2, 5)
    arctanx = el.arctan(x)
    assert(arctanx.val == pytest.approx(0.4636, 0.001))
    assert(arctanx.der == pytest.approx(2 *0.8, 0.001))
    assert(arctanx.der2 == pytest.approx(1.44, 0.01))

def test_sinh():
    assert(el.sinh(0.7) == pytest.approx(0.7585, 0.001))
    x = Dual(0.5, 2, 3)
    sinhx = el.sinh(x)
    assert(sinhx.val == pytest.approx(0.5210, 0.01))
    assert(sinhx.der == pytest.approx(2 * 1.127, 0.01))
    assert(sinhx.der2 == pytest.approx(5.467, 0.01))

def test_cosh():
    assert(el.cosh(0.6) == pytest.approx(1.185, 0.001))
    x = Dual(0.5, 2, 3)
    coshx = el.cosh(x)
    assert(coshx.val == pytest.approx(1.127, 0.001))
    assert(coshx.der == pytest.approx(2 * 0.52109, 0.001))
    assert(coshx.der2 == pytest.approx(6.073789, 0.0001))

def test_tanh():
    assert(el.tanh(0.5) == pytest.approx(0.462, 0.001))
    x = Dual(0.5, 2, 3)
    tanhx = el.tanh(x)
    assert(tanhx.val == pytest.approx(0.462, 0.001))
    assert(tanhx.der == pytest.approx(2*0.7864, 0.001))
    assert(tanhx.der2 == pytest.approx(-0.5481, 0.001))

def test_arcsinh():
    assert(el.arcsinh(0.6) == pytest.approx(0.5688, 0.001))
    x = Dual(0.5, 2, 3)
    arcsinhx = el.arcsinh(x)
    assert(arcsinhx.val == pytest.approx(0.48121, 0.001))
    assert(arcsinhx.der == pytest.approx(1.7888, 0.001))
    assert(arcsinhx.der2 == pytest.approx(1.2521, 0.001))

def test_arccosh():
    assert(el.arccosh(2) == pytest.approx(1.3169, 0.001))
    x = Dual(3, 2, 1)
    arccoshx = el.arccosh(x)
    assert(arccoshx.val == pytest.approx(1.7627, 0.001))
    assert(arccoshx.der == pytest.approx(0.3535 *2, 0.001))
    assert(arccoshx.der2 == pytest.approx(-0.17677, 0.001))

def test_arctanh():
    assert(el.arctanh(0.6) == pytest.approx(0.6931, 0.001))
    x = Dual(0.5, 2, 3)
    arctanhx = el.arctanh(x)
    assert(arctanhx.val == pytest.approx(0.5493, 0.001))
    assert(arctanhx.der == pytest.approx(1.333 *2, 0.001))
    assert(arctanhx.der2 == pytest.approx(11.11111, 0.001))

def test_exp2():
    assert(el.exp2(0) == pytest.approx(1))
    assert(el.exp2(4) == pytest.approx(16))
    x = Dual(4, 4, 1)
    x2 = el.exp2(x)
    assert(x2.val == pytest.approx(16))
    assert(x2.der == pytest.approx(44.361, 0.001))

def test_log():
    assert(el.log(27) == pytest.approx(3.2958, 0.001))
    x = Dual(3, 5, 2)
    lnx = el.log(x)
    assert(lnx.val == pytest.approx(1.098, 0.001))
    assert(lnx.der == pytest.approx(1.6666, 0.001))
    assert(lnx.der2 == pytest.approx(-2.11111, 0.001))

def test_log2():
    assert(el.log2(8) == pytest.approx(3))
    x = Dual(4, 8, 3)
    log2x = el.log2(x)
    assert(log2x.val == pytest.approx(2))
    assert(log2x.der == pytest.approx(2.8853, 0.001))
    assert(log2x.der2 == pytest.approx(-4.68875, 0.001))

def test_log10():
    assert(el.log10(100) == pytest.approx(2))
    x = Dual(10, 30, 3)
    log10x = el.log10(x)
    assert(log10x.val == pytest.approx(1))
    assert(log10x.der == pytest.approx(1.30288, 0.001))
    assert(log10x.der2 == pytest.approx(-3.778, 0.001))

def test_sqrt():
    assert(el.sqrt(5) == pytest.approx(2.236, 0.001))
    x = Dual(7, 3, 4)
    sqrtx = el.sqrt(x)
    assert(sqrtx.val == pytest.approx(2.6457, 0.001))
    assert(sqrtx.der == pytest.approx(0.5669, 0.001))
    assert(sqrtx.der2 == pytest.approx(0.63444, 0.001))

def test_sigmoid():
    assert(el.sigmoid(5) == pytest.approx(0.9933, 0.001))
    x = Dual(2, 3, 1)
    sigmoidx = el.sigmoid(x)
    assert(sigmoidx.val == pytest.approx(0.8807, 0.001))
    assert(sigmoidx.der == pytest.approx(0.10499*3, 0.001 ))

def test_relu():
    assert(el.relu(5) == 5)
    assert(el.relu(-5) == 0)
    x = Dual(2, 3, 7)
    y = Dual(-2, 3, 8)
    relux = el.relu(x)
    reluy = el.relu(y)
    assert(relux.val == 2)
    assert(relux.der == 3)
    assert(relux.der2 == 0)
    assert(reluy.val == 0)
    assert(reluy.der == 0)
    assert(reluy.der2 == 0)
